import pandas as pd
import numpy as np
from pandas._libs.tslibs import Timestamp
from typing import Tuple
from utility import data_dto

def get_number_falls(df: pd.DataFrame, start_date: Timestamp, end_date: Timestamp):
    if not df.empty:
        df = df[(df['Date'] > start_date) & (df['Date'] < end_date)]
        return len(df)
    else:
        return 0

def get_ats(ats, end_date, settings):
    screening_ats = ','.join([str(elem)[:settings['ats_delimiter']]
                              for elem in ats.DevISOClass[ats.LendDate <= end_date]])
    if screening_ats:
        return screening_ats
    return np.nan

def get_number_ats(ats, end_date):
    number_ats = ats.where(ats.LendDate <= end_date).groupby(['CitizenId'])['DevISOClass'].count()
    if number_ats.any():
        return number_ats[0]
    return 0

def get_avg_loan_period(ats, end_date):
    number_ats = ats.where(ats.LendDate <= end_date).groupby(['CitizenId'])['DevISOClass'].count()
    if number_ats.any():
        lend_ats = ats.where(ats.LendDate <= end_date)['LendDate']
        diff = lend_ats - end_date
        days = abs(diff.mean().total_seconds()) // (24 * 3600)
        return days
    return 0

def get_min_loan_period(ats, end_date):
    number_ats = ats.where(ats.LendDate <= end_date).groupby(['CitizenId'])['DevISOClass'].count()
    if number_ats.any():
        lend_ats = ats.where(ats.LendDate <= end_date)['LendDate']
        diff = lend_ats - end_date
        days = abs(diff).min().total_seconds() // (24 * 3600)
        return days
    return 0

def get_max_loan_period(ats, end_date):
    number_ats = ats.where(ats.LendDate <= end_date).groupby(['CitizenId'])['DevISOClass'].count()
    if number_ats.any():
        lend_ats = ats.where(ats.LendDate <= end_date)['LendDate']
        diff = lend_ats - end_date
        days = abs(diff).max().total_seconds() // (24 * 3600)
        return days
    return 0

def get_number_exercises(ex):
    return len([str(elem) for elem in ex.split(',') if elem != '0'])

def get_start_year(pre_screening):
    return pd.to_datetime(pre_screening.ScreeningDate).year

def get_birth_year(sc):
    return int(sc['BirthYear'].iloc[0])

def get_gender(sc):
    if (sc['Gender'].iloc[0] == 'FEMALE'):
        return 'Female'
    return 'Male'

def get_citizen_data(data: data_dto.Data, id: int) -> data_dto.Data:
    sc = data.sc.loc[data.sc['CitizenId'] == str(id)]
    ss = data.ss.loc[data.ss['CitizenId'] == str(id)]
    td = data.td.loc[data.td['CitizenId'] == str(id)]
    tc = data.tc.loc[data.tc['CitizenId'] == str(id)]
    ats = data.ats.loc[data.ats['CitizenId'] == str(id)]
    fd = data.fd.loc[data.fd['CitizenId'] == str(id)]
    citizen_data = data_dto.Data(sc, ss, td, tc, ats, fd)
    return citizen_data

def get_screening_data(td: pd.DataFrame, tc: pd.DataFrame, ss: pd.DataFrame,
                    start_date: Timestamp, end_date: Timestamp) -> Tuple[pd.DataFrame,
                                                                         pd.DataFrame,
                                                                         pd.DataFrame]:
    tdw = td.loc[(td['RatingDate'] <= end_date)
     & (td['RatingDate'] >= start_date)]
    tcw = tc.loc[(tc['RatingDate'] <= end_date)
     & (tc['RatingDate'] >= start_date)]
    ssw = ss.loc[(ss['ChangeDate'] <= end_date)
     & (ss['ChangeDate'] >= start_date)]
    return tdw, tcw, ssw

def convert_date_to_datetime(start_date, date_format):
    return pd.to_datetime(start_date, format=date_format)

def get_cancels_week(tcw):
    cancels_week = tcw['RatingDate'].apply(lambda x: "%d/%d" % (x.week, x.year))
    return cancels_week

def has_ats_id(ats, device_id):
    return any(x in ats for x in device_id)

def get_cluster(cl):
    if not cl.empty:
        return int(cl.iloc[0].Cluster)
    else:
        return 0

def get_interval_length(start_date, end_date):
    return np.around((end_date - start_date).days / 7, decimals=2)

def get_rehab_indicator(needs_start, physics_start):
    return np.around(needs_start / (physics_start + 0.0001), decimals=2)
  
def get_needs_reason(screening):
    needs_reason = screening.NeedForHelpReason
    if not needs_reason == 'nan':
        return needs_reason
    else:
        return np.nan

def get_physics_reason(screening):
    physics_reason = screening.PhysicalStrengthReason
    if not physics_reason == 'nan':
        return physics_reason
    else:
        return np.nan

def get_last_status(ssw):
    if ssw.empty:
        return np.nan
    else:
        last_status = ssw['Status'].iat[-1]
        if not pd.isnull(last_status):
            return last_status.replace(' ', '')
        else:
            return np.nan

def get_last_status_date(ssw, end_date, date_format):
    if ssw.empty:
        return end_date
    else:
        return pd.to_datetime(ssw['ChangeDate'].iat[-1], format=date_format)

def get_exercises(ex):
    return ex

def get_physics_indicator(physics_start, physics_end):
    return 1 if physics_end > physics_start else 0

def get_n_cancel_week_min(cancelsprweek):
    return cancelsprweek.value_counts().min() if not cancelsprweek.empty else 0

def get_n_weeks_without_training(n_weeks, n_weeks_with_trainings):
    return max(0, (np.ceil(n_weeks) - n_weeks_with_trainings))

def get_n_weeks_with_training(tdw, start_date):
    return tdw['RatingDate'].apply(lambda x: np.floor((x - start_date).days / 7)).nunique()

def get_training_week(tdw, start_date):
    if not tdw.empty:
        return tdw['RatingDate'].apply(lambda x: np.floor((x - start_date).days / 7))
    return pd.Series([])

def get_n_training_week_max(training_pr_week):
    if not training_pr_week.empty:
        return training_pr_week.value_counts().max()
    else:
        return 0

def get_n_training_week_min(training_pr_week, n_weeks_with_trainings, n_weeks):
    if not training_pr_week.empty and n_weeks_with_trainings > n_weeks:
        return training_pr_week.value_counts().min()
    else:
        return 0

def get_n_training_week(n_weeks, n_training_window):
    if n_training_window is not None and n_weeks is not None:
        n_trainings_week = np.around(float(n_training_window) / n_weeks, decimals=1) if n_weeks else 0
        if not pd.isnull(n_trainings_week):
            return n_trainings_week
        else:
            return 0
    else:
        return 0

def get_n_training_optimal(n_weeks):
    n_optimal = np.ceil(n_weeks * 2)
    return n_optimal

def get_n_training_window(tdw):
    n_training_window = tdw.shape[0] if not tdw.empty else 0
    return n_training_window

def get_max_evaluation(tdw):
    max_evaluation = np.around(tdw['RatingScore'].max(axis=0), decimals=1)
    if not pd.isnull(max_evaluation):
        return max_evaluation
    else:
        return 0

def get_min_evaluation(tdw):
    min_evaluation = np.around(tdw['RatingScore'].min(axis=0), decimals=1)
    if not pd.isnull(min_evaluation):
        return min_evaluation
    else:
        return 0

def get_std_evaluation(tdw):
    std_evaluation = np.around(tdw['RatingScore'].std(axis=0), decimals=1)
    if not pd.isnull(std_evaluation):
        return std_evaluation
    else:
        return 0

def get_mean_evaluation(tdw):
    mean_evaluation = np.around(tdw['RatingScore'].mean(axis=0), decimals=1)
    if not pd.isnull(mean_evaluation):
        return mean_evaluation
    else:
        return 0

def get_time_between_training_mean(tdw, n_decimals=2):
    time_between_trainings_dif = tdw['RatingDate'].diff().apply(lambda x: x.days) if not tdw.empty else None
    if time_between_trainings_dif is not None:
        if not time_between_trainings_dif.empty:
            time_mean = np.round(time_between_trainings_dif.iloc[1:].mean(), n_decimals)
            if not pd.isnull(time_mean):
                return time_mean
            else:
                return 0
        else:
            return 0
    else:
        return 0

def get_mean_time_between_cancels(tcw, n_decimals=2):
    if not tcw.empty:
        time_between_cancels_diff = tcw['RatingDate'].diff().apply(lambda x: x.days)
        if time_between_cancels_diff is not None:
            if tcw.shape[0] > 1:
                return np.round(time_between_cancels_diff.iloc[1:].mean(), n_decimals)
            else:
                return 0
        else:
            return 0
    return 0
            
def get_mean_cancels_week(n_cancel, n_weeks):
    if n_cancel is not None and n_weeks is not None:
        mean_cancels = round(float(n_cancel) / n_weeks, 2) if n_weeks else 0
        return mean_cancels
    else:
        return 0