#!/usr/bin/env python

import paths as pt
import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import GridSearchCV
from tools import file_reader, file_writer, preprocessor, data_loader, feature_maker
from kmodes import kmodes
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import StratifiedKFold
from pathlib import Path
import paths as pt
import yaml

class ClusterMaker(BaseEstimator, TransformerMixin):
    def __init__(self, init='random', n_clusters=1, n_init=1, ats_resolution=10):
        self.init = init
        self.n_clusters = n_clusters
        self.n_init = n_init
        self.ats_resolution = ats_resolution
    def fit(self, X, y=None):
        return self
    def transform(self, X, y=None):
        model = kmodes.KModes(init=self.init, n_clusters=self.n_clusters,
                              n_init=self.n_init,
                              cat_dissim=kmodes.ng_dissim,
                              n_jobs=-1)
        model.fit(X.iloc[:,-self.ats_resolution:].astype(str))
        predictions = model.predict(X.iloc[:,-self.ats_resolution:].astype(str))
        predictions = pd.Series(predictions, name="Cluster")
        X = X.iloc[: , :-self.ats_resolution].reset_index(drop=True)
        X['Cluster'] = predictions
        return X

def main():
    # Load settings
    with open(Path.joinpath(pt.CONFIGS_DIR, "settings.yaml"), 'r') as stream:
        settings = yaml.safe_load(stream)
    
    # Load screenings and make ats from them
    screenings = file_reader.read_csv(pt.INTERIM_DATA_DIR, 'screenings.csv',
                                      converters={'CitizenId': str})
    df_ats = preprocessor.split_cat_columns(screenings, col_to_split='Ats', tag='Ats', resolution=10)
    df_ats = feature_maker.make_complete_feature(df_ats, settings)
    ats_cols = df_ats.filter(regex='((\d+)[Ats])\w+', axis=1)
    
    # Load processed emb dataset
    dl = data_loader.CompleteDataLoader("complete_emb.csv", settings).load_data()
    X, y = dl.get_data()
    
    # Merge datasets
    X = pd.concat([X, ats_cols], axis=1)
    
    pipeline = Pipeline([
        ('cluster_maker', ClusterMaker()),
        ('clf', RandomForestClassifier(random_state=0, class_weight="balanced"))
    ])

    param_grid = [
        {
            'cluster_maker__init': ['random', 'Huang', 'Cao'],
            'cluster_maker__n_clusters': [5, 10, 15, 20, 30],
            'cluster_maker__n_init': [1, 5, 10, 15, 20]
        }
    ]
    
    scoring = 'average_precision'
    skf = StratifiedKFold(n_splits=5, shuffle=True, random_state=0)
    clf = GridSearchCV(pipeline, param_grid=param_grid,
                       scoring=scoring, cv=skf)
    
    clf.fit(X, y)
    
    print('\nAll results:')
    print(clf.cv_results_)
    print('\nBest estimator:')
    print(clf.best_estimator_)
    print('\nBest score:')
    print(clf.best_score_)
    print('\nBest hyperparameters:')
    print(clf.best_params_)
    results = pd.DataFrame(clf.cv_results_)
    file_writer.write_csv(results, pt.REPORTS_DIR, 'kmodes-settings-grid-search-results.csv')
    
if __name__ == '__main__':
    main()
