import os
import shutil
import numpy as np
import pandas as pd
import tensorflow as tf
import matplotlib.pyplot as plt
import os
import numpy as np
import pandas as pd
from typing import List, Tuple
from matplotlib.figure import Figure
from typing import List
from sklearn.preprocessing import LabelEncoder
from sklearn.decomposition import PCA
from pandas.api.types import is_numeric_dtype

TITLE_FORMAT = 'Weights for %s'
SCATTER_EMBEDDINGS_FORMAT = '%s_embedding.%s'
PLOT_LOSS_FORMAT = 'loss_epochs.%s'
    
def get_embedding_size(unique_values: int) -> int:
    """
    Return the embedding size to be used on the Embedding layer
    :param unique_values: the number of unique values in the given category
    :return: the size to be used on the embedding layer
    """
    size = int(min(np.ceil(unique_values / 2), 50))
    if size < 2:
        return 2
    else:
        return size
    
def get_numerical_cols(df: pd.DataFrame, target_name:str) -> List:
    """
    Generates a list of numerial categories from a dataframe
    """
    num_list = []
    for category in df:
        if not category == target_name and is_numeric_dtype(df[category]): # Only encode numerical columns
            num_list.append(category)
    return num_list

def check_weights_output(weights_output: str) -> None:
    if not weights_output:
        raise ValueError("You should provide a output file for the embeddings weights")
    
def series_to_list(series: pd.Series) -> List:
    """
    This method is used to convert a given pd.Series object into a list
    :param series: the list to be converted
    :return: the list containing all the elements from the Series object
    """
    list_cols = []
    for item in series:
        list_cols.append(item)

    return list_cols

def sample(X: np.ndarray, y: np.ndarray, n: int) -> Tuple[np.ndarray, np.ndarray]:
    num_row = X.shape[0]
    indices = np.random.randint(num_row, size=n)
    return X[indices, :], y[indices]


def get_X_y(df: pd.DataFrame, name_target: str) -> Tuple[List, List]:
    """
    This method is used to gather the X (features) and y (targets) from a given dataframe based on a given
    target name
    :param df: the dataframe to be used as source
    :param name_target: the name of the target variable
    :return: the list of features and targets
    """
    X_list = []
    y_list = []

    for index, record in df.iterrows():
        fl = series_to_list(record.drop(name_target))
        X_list.append(fl)
        y_list.append(int(record[name_target]))

    return X_list, y_list

def transpose_to_list(X: np.ndarray) -> List[np.ndarray]:
    """
    :param X: the ndarray to be used as source
    :return: a list of nd.array containing the elements from the numpy array
    """
    features_list = []
    for index in range(X.shape[1]):
        features_list.append(X[..., [index]])

    return features_list

def make_plot_from_history(history: tf.keras.callbacks.History,
                           output_path: str=None,
                           extension: str='pdf') -> Figure:
    """
    Used to make a Figure object containing the loss curve between the epochs.
    :param history: the history outputted from the model.fit method
    :param output_path: (optional) where the image will be saved
    :param extension: (optional) the extension of the file
    :return: a Figure object containing the plot
    """
    loss = history.history['loss']

    fig = plt.figure(figsize=(10, 10))
    plt.xlabel("Epochs")
    plt.ylabel("Loss")

    plt.plot(loss)

    if output_path:
        os.makedirs(output_path, exist_ok=True)
        plt.savefig(os.path.join(output_path, PLOT_LOSS_FORMAT % extension))

    return fig

def make_visualizations(labels: List[LabelEncoder],
                        embeddings: List[np.array],
                        df: pd.DataFrame,
                        output_path: str=None,
                        extension: str='pdf',
                        n_numerical_cols: int=None) -> List[Figure]:
    """
    Used to generate the embedding visualizations for each categorical variable

    :param labels: a list of the LabelEncoders of each categorical variable
    :param embeddings: a Numpy array containing the weights from the categorical variables
    :param n_numerical_cols: number of numerical columns
    :param df: the dataframe from where the weights were extracted
    :param output_path: (optional) where the visualizations will be saved
    :param extension: (optional) the extension to be used when saving the artifacts
    :param n_numerical_cols: number of numerical columns
    :return: the list of figures for each categorical variable
    """
    figures = []
    if n_numerical_cols:
        embedded_df = df.iloc[:, n_numerical_cols:df.shape[1]-1]
    else:
        embedded_df = df.iloc[:, 0:df.shape[1]-1]
        
    for index in range(embedded_df.shape[1]):
        column = embedded_df.columns[index]
        
        if is_not_single_embedding(labels[index]):
            labels_column = labels[index]
            embeddings_column = embeddings[index]

            pca = PCA(n_components=2)
            Y = pca.fit_transform(embeddings_column)

            fig = plt.figure(figsize=(10, 10))
            figures.append(fig)
            plt.scatter(Y[:, 0], Y[:, 1])
            plt.title(TITLE_FORMAT % column)
            for i, text in enumerate(labels_column.classes_):
                plt.annotate(text, (Y[i, 0], Y[i, 1]), xytext=(-20, 10), textcoords='offset points')

            if output_path:
                os.makedirs(output_path, exist_ok=True)
                plt.savefig(os.path.join(output_path, SCATTER_EMBEDDINGS_FORMAT % (column, extension)))

    return figures

def is_not_single_embedding(label: LabelEncoder) -> bool:
    """
    Used to check if there is more than one class in a given LabelEncoder
    :param label: label encoder to be checked
    :return: a boolean if the embedding contains more than one class
    """
    return label.classes_.shape[0] > 1

def get_all_columns_except(df: pd.DataFrame, columns_to_skip: List[str]) -> pd.DataFrame:
    return df.loc[:, list(filter(lambda x: x not in columns_to_skip, df.columns))]

def create_random_dataframe(rows: int=4, cols: int=4,
                            columns: str='ABCD') -> pd.DataFrame:
    return pd.DataFrame(np.random.randint(0, 10, size=(rows, cols)), columns=list(columns))

def create_random_csv(path, filename, rows: int=4, cols: int=4, columns: str='ABCD') -> str:
    df = create_random_dataframe(rows, cols, columns)
    full_path = os.path.join(path, filename)
    os.makedirs(path, exist_ok=True)
    df.to_csv(full_path, index=False)
    return full_path

def remove_random_csv(path) -> None:
    shutil.rmtree(path)